
#define PACKAGE_NAME    "Qtractor"
#define PACKAGE_VERSION "0.2.2"

#define CONFIG_PREFIX   "."
#define CONFIG_DEBUG    1

#define CONFIG_LIBVORBIS
#define CONFIG_LIBMAD
#define CONFIG_LIBSAMPLERATE

#undef  CONFIG_LIBRUBBERBAND
#undef  CONFIG_LIBLO

#define CONFIG_LADSPA
#define CONFIG_DSSI
#define CONFIG_VST

#define CONFIG_FLOAT32

#define CONFIG_GRADIENT

#undef  CONFIG_XINITTHREADS
